/*
 * JBoss, Home of Professional Open Source
 * Copyright 2012, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.jbossas.servermanager;

import com.sun.tools.attach.AttachNotSupportedException;
import com.sun.tools.attach.VirtualMachine;
import com.sun.tools.attach.VirtualMachineDescriptor;
import sun.tools.attach.HotSpotVirtualMachine;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/*
 * This class is the main class for the JStack utility. It parses its arguments
 * and decides if the command should be executed by the SA JStack tool or by
 * obtained the thread dump from a target process using the VM attach mechanism.
 *
 * Note: derived from sun.tools.jstack.JStack
 */
public class JStackLikeDump
{

   /**
    * Attach to pid and perform a thread dump.
    */
   public static String getThreadDump(final String pid) throws Exception
   {
      boolean makeDumpByKill = true;
      final StringBuffer threadDump = new StringBuffer();
      VirtualMachine vm = null;
      InputStream in = null;
      try
      {
         vm = VirtualMachine.attach(pid);

         // Cast to HotSpotVirtualMachine as this is implementation specific method.
         in = ((HotSpotVirtualMachine) vm).remoteDataDump(new String[] {"-l"});

         // read to EOF and just print output
         byte b[] = new byte[256];
         int n;
         do
         {
            n = in.read(b);
            if (n > 0)
            {
               String s = new String(b, 0, n, "UTF-8");
               threadDump.append(s);
            }
         }
         while (n > 0);

         makeDumpByKill = false;
      }
      catch (AttachNotSupportedException e)
      {
         System.err.println("The server process " + pid + " is not responding: " + e.getMessage());
         if (loadSAClass() != null)
         {
            System.err.println("trying to use Oracle's SA JStack tool instead");
            System.err.println("Check the standard output for thread dump instead of the log file.");

            runJStackTool(pid);
            makeDumpByKill = false;
         }
      }
      catch (Exception e)
      {
         System.err.println("Cannot attach to VM: " + e.getMessage());
         e.printStackTrace(System.err);
      }
      finally
      {
         if (in != null)
         {
            try
            {
               in.close();
            }
            catch (Exception ignore) {}
         }
         if (vm != null)
         {
            try
            {
               vm.detach();
            }
            catch (Exception ignore) {}
         }
      }

      if (makeDumpByKill)
      {
         System.err.println("trying to use " + killSequence + " instead (UNIX specific way)");
         threadDump.append(dumpWithKill(pid));
      }

      return threadDump.toString();
   }

   /**
    * SA JStack tool.
    */
   private static void runJStackTool(final String pid) throws Exception
   {
      Class<?> cl = loadSAClass();
      if (cl == null)
      {
         return;
      }

      // JStack tool also takes -m and -l arguments
      String params[] = new String[] {"-m", "-l", pid};

      Class[] argTypes = {String[].class};
      Method m = cl.getDeclaredMethod("main", argTypes);

      Object[] invokeArgs = {params};
      m.invoke(null, invokeArgs);
   }

   /**
    * @return Returns sun.jvm.hotspot.tools.JStack if available, otherwise null.
    */
   private static Class loadSAClass()
   {
      //
      // Attempt to load JStack class - we specify the system class
      // loader so as to cater for development environments where
      // this class is on the boot class path but sa-jdi.jar is on
      // the system class path. Once the JDK is deployed then both
      // tools.jar and sa-jdi.jar are on the system class path.
      //
      try
      {
         return Class.forName("sun.jvm.hotspot.tools.JStack");
      }
      catch (Exception ignore) {}
      return null;
   }

   /**
    * Returns a set of PIDs of all running JVMs which match given mainClass.
    */
   public static Set<String> getJVMPids(final String mainClass)
   {
      final Set<String> jvmPids = new HashSet<String>();

      // go through all running JVMs and take those which contain AS's mainClass
      final List<VirtualMachineDescriptor> vms = VirtualMachine.list();
      for (VirtualMachineDescriptor vm : vms)
      {
         if (vm.displayName().indexOf(mainClass) >= 0)
         {
            jvmPids.add(vm.id());
         }
      }

      return jvmPids;
   }


   private static String killSequence = "kill -3 [pid]";

   protected static String dumpWithKill(final String pid)
   {
      final StringBuffer killOutput = new StringBuffer();

      final String os = System.getProperty("os.name").toLowerCase();
      if (os.indexOf("windows") > -1)
      {
         // unsupported on MS Windows
         return killOutput.append("Kill command unsupported on this platform").toString();
      }

      final String killCommand = killSequence.replace("[pid]", pid);

      final String nl = System.getProperty("line.separator");
      killOutput.append("calling " + killCommand + " ...").append(nl);

      Process p = null;
      InputStream out = null;
      try
      {
         p = Runtime.getRuntime().exec(killCommand);

         p.waitFor();
         int rc = p.exitValue();

         if (rc != 0)
         {
            killOutput.append("Kill sequence failed").append(nl);
         }

         out = p.getInputStream();
         BufferedReader in = new BufferedReader(new InputStreamReader(out));
         String result;
         do
         {
            result = in.readLine();
            if (result != null) killOutput.append(result).append(nl);
         }
         while (result != null);

      }
      catch (Exception e)
      {
         killOutput.append("Kill sequence failed: ").append(e.getMessage());
      }
      finally
      {
         if (out != null)
         {
            try
            {
               out.close();
            }
            catch (IOException ignore) {}
         }
         if (p != null)
         {
            p.destroy();
         }
      }

      return killOutput.toString();
   }

   // For debugging purposes only
   public static void main(String[] argv)
   {
      System.out.println("JStackLikeDump.loadSAClass()=" + JStackLikeDump.loadSAClass());

      Set<String> jvmPIDs = JStackLikeDump.getJVMPids("org.jboss.Main");
      System.out.println("jvmPIDs = " + jvmPIDs);
      for (String pid : jvmPIDs)
      {
         try
         {
            System.out.println("===== Thread dump of " + pid + " PID");
            System.out.println(JStackLikeDump.getThreadDump(pid));
            System.out.println("=====");
         }
         catch (Exception e)
         {
            System.err.println("Unable to get server thread dump of " + pid + " PID!");
            e.printStackTrace(System.err);
         }
      }
   }

}
